//
// =======================================================================================================
// IMPORTANT, PLEASE READ!!
// =======================================================================================================
//

/* To allow displaying of "°":
- In the Arduino IDE, go to Files -> Preferences
- At the bottom of the Preference window you have a link to the preferences.txt file
- Click it to open the file, then close the IDE
- In preferences.txt, find setting preproc.substitute_unicode and change it's value from true to false


To disable the limit of 100 EEPROM write operations:
remove the "_EEPROMEX_DEBUG" #define in the EEPROMex library source code!
*/

//
// =======================================================================================================
// BUILD OPTIONS (comment out unneeded options)
// =======================================================================================================
//

#define CAPACITIVE // support for capacitive touch sensors (for inverter brightness adjustment)
//#define SERVO // support for an RC Servo (intended for a door, which covers front USB ports)

//
// =======================================================================================================
// LIRBARIES
// =======================================================================================================
//
#include <OneWire.h>
#include <DallasTemperature.h>
#include <SimpleTimer.h>
#include <EEPROMex.h>  // NOTE: Please remove the "_EEPROMEX_DEBUG" #define in the library source code!
#include <PWMFrequency.h>

#ifdef CAPACITIVE
#include <CapacitiveSensor.h>
#endif

#ifdef SERVO
#include <SoftwareServo.h>    // Important! Servo judders with Servo.h
#endif

// The Libraries PWMFrequency & SimpleTimer can be obtained here:
// https://github.com/kiwisincebirth/Arduino/tree/master/Libraries

//
// =======================================================================================================
// VERSION COMMENTS
// =======================================================================================================
//
// This sketch is for my iMac G5 A1058 Hackintosh Mod and provides almost exactly the same system behavior as in original condition:
// - Front LED breathing during Sleep (S3) or always on, if System is on (S0) and CCFL is off
// - CCFL Backlight Brightness adjustment via Slider Applet or Capacitive Touch Sensors
// - Apple Chime Start Sound via original Speakers (The Speakers and the AMP are shared with the system USB sound device)
// - Original PSU Controller
// - Original Fan Controller (PWM controlled in accordance with three Temperature Sensors
// - RC Servo Support (not used in this Build, intended for pushing a iMac G4 front door open)
// http://www.tonymacx86.com/imac-mods/130197-mactester57s-imac-g5-17-a1058-model.html
//
// Parts of this sketch are based on the excellent work of kiwisincebirth:
// http://www.tonymacx86.com/imac-mods/107859-kiwis-next-project-imac-g5.html
// https://github.com/kiwisincebirth/iMacG5
//
// V.0.9 Jan. 2. 2015 MacTester (Working with NUC-Dummy inside iMac G5)
// V.0.9.1 enhanced PSU shutdown delay from 20 to 30s and reduced PS impulse from 1 to 0.3s in order to solve weird hang issues
// V.0.9.2 fixed inverter value mapping to 0-100, 140 - 0, PSU ts = millis(), if PS pressed...
// V.0.9.4 State detection: S5, if LOW, LOW. Inverter startup brightness 70%. PS impulse 0.7s -> Runs OK!
// V.0.9.5 state = 1, if power switch pressed
// V.0.9.5.2 actualised for using the old SMC prototype board (old pinout) with the new Brightness menulet (terminal functionality etc)
// V.0.9.6 Touch sensors 1M resistors instead of 10M, Fan PWM frequency reduced from 31.5kHz to 123Hz (original iMac =100Hz)
// V.0.9.7 Reworked for new board "iMac SMC V1.0" with different pinout. Fan RPM measurement with interrupts, WTV020-DS-16P Chime module, PAM8403 amplifier.
// V.0.9.8 Eeprom functionality added, custom PWMFrequency library replaced with PWMFrequency.h, power state detection rewritten
// V.0.9.9 First Beta Release, ready to install the new "SMC V1.0" Board in the iMac
// V.1.0.0 CPU Fan stays completely off below 42 deg C of the NUC exhaust air (more silent). EEPROM Fan defaults changed. PS Impulse 0.6s.

//
// =======================================================================================================
// PIN ASSIGNMENTS & GLOBAL VARIABLES
// =======================================================================================================
//

//
// Digital 0, 1, Shared as serial pins (Serial1)
// Digital 2 - 13, General Purpose
// Digital 3, 5, 6, 9, 10, 11, and 13 - PWM
// Digital 11 - 13 Excluded on Pro Micro
// Digital 13 has (not Leo) an LED and resistor attached to it
// Digital 14 - 16, Shared with an SPI bus
// Digital 17 - Doubles and RX LED
// Digital Pins can sink 40ma, recommended 470 ohm resistor
// Analog Inputs: A0-A3, Pro Micro excludes A4, A5
// Analog Inputs: A6-A11 (on digital pins 4, 6, 8, 9, 10, and 12)
// Analog Pins can be used as digital ( refer to them as A0, A1, ...)
//

//----PSU CONTROLLER
#define PSU_DEACTIVATE_POUT 4 // if HIGH, the main PSU will be OFF -> (only 5V Standby is available)

//----FAN CONTROLLER

// Temperature Sensor Pin & other parameters
#define ONE_WIRE_BUS A2 //One Wire BUS is Pin 16 (all Temperature Sensors are wired there)
OneWire oneWire(ONE_WIRE_BUS); //Setup a oneWire instance to communicate with OneWire Devices
DallasTemperature sensors(&oneWire); //Pass our oneWire reference to Dallas Temperature
DeviceAddress tempDeviceAddress0;
DeviceAddress tempDeviceAddress1;
DeviceAddress tempDeviceAddress2;
byte  resolution = 10; //Sensor resolution = 10bit = 0.25°C increment (9 to 12bit possible)

//Fan PWM Output
#define FAN1_CONTROL_POUT 9 // PWM Fan Speed Output (Case Fan)
#define FAN2_CONTROL_POUT 6 // PWM Fan Speed Output (HDD Fan)
#define FAN3_CONTROL_POUT 10 // PWM Fan Speed Output (CPU Fan)

//Fan Tach Inputs
#define FAN1_TACH_INT 1 // Fan1 Speed Tachometer Input
#define FAN2_TACH_INT 2 // Fan2 Speed Tachometer Input
#define FAN3_TACH_INT 7 // Fan3 Speed Tachometer Input

//----LED CONTROLLER, POWER SWITCH & NUC CS HEADER
#define SLEEP_LED_POUT 3    // LED connected to digital pin 3
#define CASE_POWER_SWITCH_PIN A1  // Case Switch
#define NUC_POWER_LED_CS_PIN2 14  // NUC Custom Solution Header Pin 2
#define NUC_POWER_LED_CS_PIN4 15  // NUC Custom Solution Header Pin 4

//----LCD INVERTER CONTROLLER
#define INVERTER_POUT 5    // CCFL inverter
boolean inverterEnabled;

//----SERVO CONTROLLER
#define DOOR_SERVO_POUT 0    // Door servo connected to digital pin 0 RXI (a PWM pin is not required for servos)
#ifdef SERVO
SoftwareServo doorServo;
#endif

//----CAPACITIVE TOUCH SENSORS
// Up and down capacitive touch brightness pins
#define TOUCH_COMMON_PIN_CAP A0
#define TOUCH_DOWN_PIN_CAP 16
#define TOUCH_UP_PIN_CAP 8

#ifdef CAPACITIVE
// 1M resistor between pins 2 & 8, pin 8 is sensor pin
CapacitiveSensor dnSensor = CapacitiveSensor(TOUCH_COMMON_PIN_CAP, TOUCH_DOWN_PIN_CAP);

// 1M resistor between pins 2 & 7, pin 7 is sensor pin
CapacitiveSensor upSensor = CapacitiveSensor(TOUCH_COMMON_PIN_CAP, TOUCH_UP_PIN_CAP);
#endif

//----APPLE CHIME START MELODY
#define CHIME_POUT A3

//----System diagnosis
unsigned long millisOfCpuStartup = millis();
unsigned long millisOfEnteringSleep = millis();
unsigned long millisWhenPSUActivated = millis();

//----TIMER OBJECT
SimpleTimer timer;

//----Some ASCII Constants
#define CR 13
#define LF 10
#define TAB '\t'
#define STOP ','
#define COMMA ','
#define COLON ':'
#define EQUALS '='
#define ZERO '0'
#define RPM " RPM   "
#define DEGC " deg C " //" °C" // see comment on top of the sketch! BUT, "°" does not work in the Slider Applet

#define SERIAL_BAUD_RATE 115200

//
// =======================================================================================================
// EEPROM Locations that configuration data is stored
// =======================================================================================================
//

// Location the Current Version ID is stored
#define EEP_VERSION 0

// retentive process output variables
#define EEP_BRIGHT 1 // brightness Setting 0 - 100 %
#define EEP_CYC_MAINS 2 // SMC Power / Reset Cycles (StartUp)
#define EEP_CYC_POWER 3 // CPU Power Cycles (Power Up)
#define EEP_CYC_SLEEP 4 // CPU Sleep Cycles (Wake)
#define EEP_TIM_PSU 5 // PSU Uptime
#define EEP_TIM_LCD 6 // LCD Backlight Power On Time
#define EEP_TIM_SLEEP 7 // Sleep (S3) Time

#define EEP_RESERVED3 8
#define EEP_RESERVED4 9
#define EEP_RESERVED5 10
#define EEP_RESERVED6 11
#define EEP_RESERVED7 12
#define EEP_RESERVED8 13
#define EEP_RESERVED9 14
#define EEP_RESERVEDA 15

// retentive process input variables (displayed with "ER" command)
#define EEP_MIN_BRIGHT 16 // Minimum PWM for Inverter - When Brightness is 0%
#define EEP_MAX_BRIGHT 17 // Maximum PWM for Inverter - When Brightness is 100% (0 PWM = full brightness)
#define EEP_BRIGHT_INC 18 // Number of percent brightness Increased or decreased
#define EEP_MIN_FAN1_PWM 19 // Fan 1 Control min PWM
#define EEP_FAN1_DIV 20 // Fan 1 Control Divider
#define EEP_MIN_FAN2_PWM 21 // Fan 2 Control min PWM
#define EEP_FAN2_DIV 22 // Fan 2 Control Divider
#define EEP_MIN_FAN3_PWM 23 // Fan 3 Control min PWM
#define EEP_FAN3_DIV 24 // Fan 3 Control Divider
#define EEP_FAN3_TRESH 25  // The CPU Fan will not be actice below this temperature
#define EEP_MIN_LED_PWM 26 // Minimum PWM Value for Front Panel LED Effects
#define EEP_MAX_LED_PWM 27 // Maximum PWM Value for Front Panel LED Effects
#define EEP_DN_CAP_THR 28 // Capacitance Threshold for Down Button
#define EEP_UP_CAP_THR 29 // Capacitance Threshold for Up Button
#define EEP_CAP_SAMPLE 30 // Number of Samples for Capacitance
#define EEP_INVERTER_STARTUP_DELAY 31 // milliseconds before starting inverter 
#define EEP_CHIME_STARTUP_DELAY 32 // milliseconds before triggering the Apple Chime
#define EEP_CHIME_DURATION 33  // the fans stay off during this time, if the Apple Chime was triggered
#define EEP_DOOR_POS_CLOSED 34 // the closed position of the door
#define EEP_DOOR_POS_OPENED 35 // the opened position of the door
#define EEP_DOOR_SPEED 36 // the door speed

//
// Names of Data stored in these locations (max. 64 charatcers, including values due to buffer size in eepromName()!! )
//

const char PROGMEM string_RES[] PROGMEM = "Reserved";
const char PROGMEM string_DEP[] PROGMEM = "Deprecated";
const char PROGMEM string_VS[] PROGMEM = "Vers.-Size";

const char PROGMEM string_P0[] PROGMEM = "Bright-%";  // 1
const char PROGMEM string_P1[] PROGMEM = "Cycles-SMC";  // 2
const char PROGMEM string_P2[] PROGMEM = "Cycles-Power";  // 3
const char PROGMEM string_P3[] PROGMEM = "Cycles-Sleep";  // 4
const char PROGMEM string_P4[] PROGMEM = "Time-Power";  // 5
const char PROGMEM string_P5[] PROGMEM = "Time-LCD";  // 6
const char PROGMEM string_P6[] PROGMEM = "Time-Sleep";  // 7

const char PROGMEM string_00[] PROGMEM = "CCFL Brightness Min. PWM"; // 16
const char PROGMEM string_01[] PROGMEM = "CCFL Brightness Max. PWM";
const char PROGMEM string_02[] PROGMEM = "CCFL Brightness Inc. %";
const char PROGMEM string_03[] PROGMEM = "FAN 1 (Case) Min. PWM";
const char PROGMEM string_04[] PROGMEM = "FAN 1 Divider";
const char PROGMEM string_05[] PROGMEM = "FAN 2 (HDD) Min. PWM";
const char PROGMEM string_06[] PROGMEM = "FAN 2 Divider";
const char PROGMEM string_07[] PROGMEM = "FAN 3 (CPU) Min. PWM";
const char PROGMEM string_08[] PROGMEM = "FAN 3 Divider";
const char PROGMEM string_09[] PROGMEM = "FAN 3 act. threshold deg C";
const char PROGMEM string_10[] PROGMEM = "LED Min. PWM";
const char PROGMEM string_11[] PROGMEM = "LED Max. PWM"; // 27
const char PROGMEM string_12[] PROGMEM = "Cap. Touch Down Threshold";
const char PROGMEM string_13[] PROGMEM = "Cap. Touch Up Threshold";
const char PROGMEM string_14[] PROGMEM = "Cap. Touch Samples";
const char PROGMEM string_15[] PROGMEM = "Inverter Start Delay";
const char PROGMEM string_16[] PROGMEM = "Chime Start Delay";
const char PROGMEM string_17[] PROGMEM = "Chime Duration";
const char PROGMEM string_18[] PROGMEM = "Door Pos. closed";
const char PROGMEM string_19[] PROGMEM = "Door Pos. opened";
const char PROGMEM string_20[] PROGMEM = "Door Speed"; // 36

//
// the following must contain all strings from above
//

// PGM_P is used to declare a variable that is a pointer to a string in program space.

PGM_P const eepName[] PROGMEM = {
  string_VS,

  string_P0, string_P1, string_P2, string_P3, string_P4, string_P5, string_P6,  // 1 - 7

  string_RES, string_RES, string_RES, string_RES, string_RES, string_RES, string_RES, string_RES,  // 8 - 15

  string_00, string_01, string_02, string_03, string_04, // 16 - 35
  string_05, string_06, string_07, string_08, string_09,
  string_10, string_11, string_12, string_13, string_14,
  string_15, string_16, string_17, string_18, string_19, string_20,
};

//
// =======================================================================================================
// MAIN ARDUINO SETUP (1x during startup)
// =======================================================================================================
//

void setup() {

  // Start up the One Wire Sensor Library
  sensors.begin();
  sensors.getAddress(tempDeviceAddress0, 0);
  sensors.getAddress(tempDeviceAddress1, 1);
  sensors.getAddress(tempDeviceAddress2, 2);
  sensors.setResolution(tempDeviceAddress0, resolution);
  sensors.setResolution(tempDeviceAddress1, resolution);
  sensors.setResolution(tempDeviceAddress2, resolution);

  //Set PinModes

  pinMode(PSU_DEACTIVATE_POUT, OUTPUT);
  digitalWrite(PSU_DEACTIVATE_POUT, LOW); //PSU on after mains was switched on!

  pinMode(FAN1_CONTROL_POUT, OUTPUT);
  pinMode(FAN2_CONTROL_POUT, OUTPUT);
  pinMode(FAN3_CONTROL_POUT, OUTPUT);
  digitalWrite(FAN1_CONTROL_POUT, LOW);
  digitalWrite(FAN2_CONTROL_POUT, LOW);
  digitalWrite(FAN3_CONTROL_POUT, LOW);

  pinMode(FAN1_TACH_INT, INPUT_PULLUP);
  pinMode(FAN2_TACH_INT, INPUT_PULLUP);
  pinMode(FAN3_TACH_INT, INPUT_PULLUP);

  pinMode(SLEEP_LED_POUT, OUTPUT);
  digitalWrite(SLEEP_LED_POUT, LOW);

  pinMode(CASE_POWER_SWITCH_PIN, INPUT_PULLUP);
  pinMode(NUC_POWER_LED_CS_PIN2, INPUT_PULLUP);
  pinMode(NUC_POWER_LED_CS_PIN4, INPUT_PULLUP);

  pinMode(CHIME_POUT, OUTPUT);
  digitalWrite(CHIME_POUT, HIGH);

  pinMode(INVERTER_POUT, OUTPUT);
  digitalWrite(INVERTER_POUT, HIGH);

  //Attach servo to Pin
#ifdef SERVO
  doorServo.attach(DOOR_SERVO_POUT);
#endif

  // Timer, which triggers the sleep LED brightness attenuator every 15ms
  timer.setInterval(15, ledAttenuator);

#ifdef CAPACITIVE
  // Timer, which triggers the touch sensor reading every 182ms
  timer.setInterval(182, touchControl);
#endif

#ifdef SERVO
  // Timer, which triggers the door servo function every 20ms
  timer.setInterval(20, servoFunction);
  doorServo.write(eepromRead(EEP_DOOR_POS_CLOSED));  // go to the servo initializing position
#endif

  // Adjust PWM frequencies
  // Note: the original iMac G5 fans are driven @ 100Hz
  setPWMPrescaler(FAN1_CONTROL_POUT, 256); // Sets 123Hz Frequency ( 256 )
  setPWMPrescaler(FAN2_CONTROL_POUT, 256); // Sets 123Hz Frequency ( 256 )
  setPWMPrescaler(FAN3_CONTROL_POUT, 256); // Sets 123Hz Frequency ( 256 )
  
  // Enable inverting mode for PWM Output 9 + 10
  //TCCR1A = TCCR1A | 0x30;

  setPWMPrescaler(INVERTER_POUT, 1); // Sets 31.25KHz Frequency ( 1 )

  // Fan tachometer interrupt initialisation
  attachInterrupt(3, interrupFan1, FALLING); // pin 1 (TX0) = interrupt 3
  attachInterrupt(1, interrupFan2, RISING); // pin 2 = interrupt 1
  attachInterrupt(4, interrupFan3, FALLING); // pin 7 = interrupt 4

  // Inverter setup
  setupInverter();

  // EEPROM configuration
  checkEepromConfiguration();

#ifdef CAPACITIVE
  // Calibrate the touch sensors
  delay (500);                // Wait 0.5s...
  setupTouch();               // ... then initialize the touch sensors
#endif

  // count mains power cycles
  incrementCounter(EEP_CYC_MAINS);
}

//
// =======================================================================================================
// MAIN LOOP
// =======================================================================================================
//

void loop() {

  // Loop the timer
  timer.run();

  // Loop the powerState detection controller
  powerStat();

  // Loop the Power Switch FSM
  fsmPowerSwitch();

  // Loop the PSU Controller
  psuController();

  // Loop the Chime FSM
  fsmChime();

  // Loop the Fan Controller FSM
  fsmFans();

  // Loop the LED Controller
  ledSinus();

  // Loop the inverter Controller
  inverterOnOff();

#ifdef SERVO
  // Refresh the servo position
  SoftwareServo::refresh();
#endif

  // Loop USB serial controller
  loopSerialCommandProcessor();
}

//
// =======================================================================================================
// EEPRORM SETUP AND USER CONFIGURATION
// =======================================================================================================
//

int epromWriteCnt = 0; // # of EEPROM write accesses since last SMC reboot

/**
 * Setup the device if not initialised, flasing the eeprom with defaults, if they don't exist
 */
void checkEepromConfiguration() {

  // The version number holds the number of the highest value stored in eeprom
  // This allows future software upgrades with new eeprom values, to have defaults
  // written to the eeprom without overitting exisitng values that are stored there
  // VERSION = The Numer of (the highest) items that already have values, dont overitte

  // Read the current version of the eeprom
  byte currentHighest = eepromRead(EEP_VERSION);
  if ( currentHighest == 255 ) currentHighest = 0;

  // All The Default Values
  if ( currentHighest < EEP_DOOR_SPEED ) {

    // write default values
    eepromWrite( EEP_BRIGHT, 70 ); // 70% brightness
    eepromWrite( EEP_CYC_MAINS, 0);   //
    eepromWrite( EEP_CYC_POWER, 0); //
    eepromWrite( EEP_CYC_SLEEP, 0); //
    eepromWrite( EEP_TIM_PSU, 0); //
    eepromWrite( EEP_TIM_LCD, 0); //
    eepromWrite( EEP_TIM_SLEEP, 0); //
    eepromWrite( EEP_RESERVED3, 0); //
    eepromWrite( EEP_RESERVED4, 0); //
    eepromWrite( EEP_RESERVED5, 0); //
    eepromWrite( EEP_RESERVED6, 0); //
    eepromWrite( EEP_RESERVED7, 0); //
    eepromWrite( EEP_RESERVED8, 0); //
    eepromWrite( EEP_RESERVED9, 0); //
    eepromWrite( EEP_RESERVEDA, 0); //

    eepromWrite( EEP_MIN_BRIGHT, 140 ); // ca. 140 PWM = low brightness, but CCFL stays on!
    eepromWrite( EEP_MAX_BRIGHT, 0 ); // 0 PWM = full brightness!!
    eepromWrite( EEP_BRIGHT_INC, 5 ); // Brightness Increment = 5
    eepromWrite( EEP_MIN_FAN1_PWM, 20 ); // Minimum PWM value for Fan 1 (Case) = 20
    eepromWrite( EEP_FAN1_DIV, 2000 ); // Fan 1 Divider = 2000
    eepromWrite( EEP_MIN_FAN2_PWM, 20 ); // Minimum PWM value for Fan 2 (HDD) = 20
    eepromWrite( EEP_FAN2_DIV, 1400 ); // Fan 2 Divider = 1400
    eepromWrite( EEP_MIN_FAN3_PWM, 16 ); // Minimum PWM value for Fan 3 (CPU) = 16
    eepromWrite( EEP_FAN3_DIV, 2500 ); // Fan 3 Divider = 2500
    eepromWrite( EEP_FAN3_TRESH, 42 ); // Fan 3 will be active beyond 42 deg C
    eepromWrite( EEP_MIN_LED_PWM, 14 ); // Minimum PWM for Breathing LED = 14
    eepromWrite( EEP_MAX_LED_PWM, 255 ); // Maximum Brightness for breathing LED = 255
    eepromWrite( EEP_DN_CAP_THR, 70 ); // Down Threshhold = 70
    eepromWrite( EEP_UP_CAP_THR, 70 ); // Up Threshold = 70
    eepromWrite( EEP_CAP_SAMPLE, 80 ); // Samples = 80
    eepromWrite( EEP_INVERTER_STARTUP_DELAY, 6500 ); // Inverter Startup delay miliseconds = 6500
    eepromWrite( EEP_CHIME_STARTUP_DELAY, 2500 ); // Apple Chime Delay = 2500
    eepromWrite( EEP_CHIME_DURATION, 2800 ); // Apple Chime Duration = 2800
    eepromWrite( EEP_DOOR_POS_CLOSED, 0 ); // Closed position of the door servo = 0
    eepromWrite( EEP_DOOR_POS_OPENED, 180 ); // Opened position of the door servo = 180
    eepromWrite( EEP_DOOR_SPEED, 3 ); // The door servo speed is between 1 and 6

    // EEPROM number of entries (latest entry from above)
    currentHighest = EEP_DOOR_SPEED;
  }

  // finally update eprom with Highest numbered item stored in EEPROM
  eepromWrite(EEP_VERSION, currentHighest);
}

//----------------------------- PUBLIC

long eepromRead(byte location) {
  // Can write to an invalid location
  if (isLocationException(location) ) return -1;
  return EEPROM.readLong( eepromLocation(location) );
}

int eepromWrite(byte location, long value) {

  // Can write to an invalid location
  if (isLocationException(location) ) return -1;

  // the eeprom library protects from writing same values
  epromWriteCnt++;
  return EEPROM.updateLong( eepromLocation(location), value );
}

void inline incrementCounter(byte location) {
  addToCounter(location, 1);
}

void addToCounter(byte location, unsigned long value) {
  eepromWrite(location, eepromRead(location) + value);
}

// ------------------ PRIVATE

byte sizeofEeprom() {
  return EEP_DOOR_SPEED + 1; // HARD CODED
  //return eepromRead(EEP_VERSION) + 1; // does not work!
}

boolean isLocationException(byte location) {
  if (location >= sizeofEeprom() ) {
    return true;
  }
  return false;
}

// eeprom addresses <16 are write protected!
boolean isProtectedException(byte location) {
  return location < 16;
}

int inline eepromLocation(byte location) {
  return location * 4;
}

String eepromName(byte location) {
  static char buffer[64];  // Max. length of allowed EEPROM entry comments (including variables)!!
  if (isLocationException(location) ) return "";
  strcpy_P(buffer, (char*)pgm_read_word(&(eepName[location])) );
  return buffer;
}

// -------------------------------

/**
 * Process a command from the serial port
 */
void processCommandEeprom(String subCmd, String extraCmd) {

  if (subCmd.equals("R")) {
    // EEPROM Read
    eepromReadCommand(extraCmd);

  } else if (subCmd.equals("W")) {
    // EEPROM Write EWnnn,xxxxx where n is the address and xxxx is the value
    eepromWriteCommand(extraCmd);

  } else if (subCmd.equals("F")) {
    // Reset The eeprom firmware version and reinit.
    eepromInitCommand();

  } else {
    Serial.println(F("EEPROM Command Unknown: ERn (read), EWn,v (write), EF (factory)"));
  }
}

/**
 * The Read EEPROM command can be passed with an optional argument
 * n - which specifies the location to read, rather than iterating all locations.
 */
void eepromReadCommand(String extraCmd) {

  // if parameter passed then convert
  byte location = extraCmd.toInt();
  if (location > 0 && !isLocationException(location) ) {
    eepromPrintCommand(location);
  } else {
    for (byte i = 16; i < sizeofEeprom(); i++) { // start @ entry 16
      eepromPrintCommand(i);
    }
  }
}

void eepromPrintCommand(byte location) {

  Serial.print(location);
  Serial.print(TAB);

  Serial.print(eepromName(location));
  Serial.print(TAB);

  Serial.println(eepromRead(location));
}

/**
 * The Write command has the following syntax
 * EWn,v - e,g EW7,3000  - writes the value of 3000 to address 7
 * Note: Some locations are protected, so cannot be written
 */
void eepromWriteCommand(String extraCmd) {

  // look for comma.
  byte index = extraCmd.indexOf(COMMA);
  if (index <= 0) {
    // no equal so cannot handle
    Serial.println(F("Syntax Error - Correct Syntax is: EWn,v "));
    Serial.println(F("n is the address, v is the value"));
  } else {

    // work out the item and the location
    String loc = extraCmd.substring(0, index);
    String val = extraCmd.substring(index + 1);
    byte location = loc.toInt();
    long value = val.toInt();

    if ( isLocationException(location) || isProtectedException(location) ) {
      Serial.print(F("Cannot write to address "));
      Serial.println(location);
      return;
    }

    // write to eprom obeying protected status
    eepromWrite(location, value);
    Serial.println(F("EEPROM entry flashed. -> Check with ER"));
  }
}

/**
 * The Flash command is used to re-initialise the firmware
 * Normally this shouldn't be needed. Only use if you need to recover defaults.
 */
void eepromInitCommand() {

  // EPROM Initialise. Resets the Eprom Values to default Values, suggest reboot after this
  eepromWrite(EEP_VERSION, 0);

  // forces the eeprom to be re-inited
  checkEepromConfiguration();

  Serial.println(F("EEPROM defaults restored."));
}

//
// =======================================================================================================
// NUC POWER STATE DETECTION
// =======================================================================================================
//

byte powerState; // global variable, holding the system power state

byte stateDetect() {

  // This function reads the "raw" NUC power state
  byte rawPowerState;

  // read voltages on two CS header pins
  boolean POS_Val = digitalRead(NUC_POWER_LED_CS_PIN2);
  boolean Neg_Val = digitalRead(NUC_POWER_LED_CS_PIN4);

  // detect the NUC power state in accordance with pin voltages
  if ( POS_Val == HIGH && Neg_Val == LOW  ) rawPowerState = 0;  // S0 = All ON
  if ( POS_Val == LOW  && Neg_Val == HIGH ) rawPowerState = 3;  // S3 = Standby Mode (PSU has to stay ON, because the NUC still requires 12V)
  if ( POS_Val == HIGH && Neg_Val == HIGH ) rawPowerState = 5;  // S5 = Soft Off (only 5V standby is ON, all other supply voltages are OFF)
  if ( POS_Val == LOW && Neg_Val == LOW ) rawPowerState = 5;

  return rawPowerState;
}

// This function debounces the system power state. Reason: the NUC power state is not 100% reliable and has sometimes some "spikes" in it...
// Also the sleep wake cycles are counted & the duration of sleep phases is added to the time in the EEPROM
void powerStat() {

  static unsigned long changedWhen; // the time, when the raw signal changed
  static byte lastPowerState; // internal variable, used to detect raw state change
  static boolean wasInS3;
  static boolean wasInS0;
  static boolean wasNotInS5;

  // if the state is different (due to noise or signal change)
  if ( stateDetect() != lastPowerState ) {
    changedWhen = millis();  // save, when the state did change
    lastPowerState = stateDetect();
  }

  //if the debounce time (1s) has passed, report the new power state
  if (millis() - changedWhen > 1000) {
    powerState = stateDetect();
  }

  // Waking from sleep
  if (powerState == 3) wasInS3 = true;
  if (powerState == 0 && wasInS3) {
    incrementCounter(EEP_CYC_SLEEP); // count wake from sleep cycles
    addToCounter(EEP_TIM_SLEEP, (millis() - millisOfEnteringSleep) / 1000L); // add (millis - the wake time) to the time in the EEPROM
    wasInS3 = false;
  }

  // Going to sleep
  if (powerState == 0) wasInS0 = true;
  if (powerState == 3 && wasInS0) {
    millisOfEnteringSleep = millis(); // save, when the state did change to S3
    wasInS0 = false;
  }

  // Shutdown
  if (powerState != 5) wasNotInS5 = true;
  if (powerState == 5 && wasNotInS5) {
    writeBrightToEEPROM();  // save the current inverter brightness to the EEPROM
    wasNotInS5 = false;
  }

  // detect, if not in S0 (ON) - for System Uptime SU
  if (powerState != 0) millisOfCpuStartup = millis();  // Save the time, when the CPU was activated
}

//
// =======================================================================================================
// POWER SWITCH FSM (Finite State Machine)
// =======================================================================================================
//

void fsmPowerSwitch() {

  static byte state = 1; // This variable is holding the current "step" of the FSM
  static unsigned long ts;  // To store the "current" time in for delays.
  boolean switchVal = digitalRead(CASE_POWER_SWITCH_PIN);  //Read switch

  switch (state) {

    //----Is Power Switch pressed? -> Generate a 0.6s "LOW-Pulse", even if the power swich button was released before!
    //    Reason: The PSU must be enabled before the NUC can be switched on!
    case 1:
      if (switchVal == false) {
        pinMode(CASE_POWER_SWITCH_PIN, OUTPUT);  //Make it an Output...
        digitalWrite(CASE_POWER_SWITCH_PIN, LOW);  //And pull it LOW for 0.6s
        state = 2;
      }
      else state = 4;  // Else don't make Pin an output
      break;

    //----Start timer
    case 2:
      ts = millis();  // Remember the current time
      state = 3;
      break;

    //----If 0.6s has passed...
    case 3:
      if (millis() > ts + 600) {
        state = 4;
      }
      break;

    //----Make power switch Pin an input again
    case 4:
      pinMode(CASE_POWER_SWITCH_PIN, INPUT_PULLUP);
      state = 1;
      break;

    //----
    default:
      state = 1;
      break;
  }
}

//
// =======================================================================================================
// MAIN PSU CONTROLLER
// =======================================================================================================
//

void psuController() {

  boolean switchVal = digitalRead(CASE_POWER_SWITCH_PIN);  //Read switch
  boolean psuIsDisabled = digitalRead(PSU_DEACTIVATE_POUT); //Read state of PSU deactivation output
  static boolean shutoffTimeSaved = false;
  static unsigned long shutoffMillis;

  // Enable the main PSU, if the power switch is pressed & the PSU is not yet enabled!
  if (switchVal == false && psuIsDisabled) { //false = Switch pressed!
    digitalWrite(PSU_DEACTIVATE_POUT, LOW); // LOW = PSU enabled
    //psuIsDisabled = true;
    millisWhenPSUActivated = millis(); // reset counter
    shutoffTimeSaved = false;
  }

  // remember the time, when the computer changes to S5 (OFF)
  if (powerState == 5 && ! shutoffTimeSaved) {
    shutoffMillis = millis();
    shutoffTimeSaved = true;
  }

  // allow to save the shutoff-time again, if the computer is NOT in S5 (OFF) state
  if (powerState != 5) {
    shutoffTimeSaved = false;
  }

  // Disable the main PSU after 5s in S5 (OFF) state
  if (powerState == 5 && millis() - shutoffMillis > 5000 && ! psuIsDisabled ) {
    digitalWrite(PSU_DEACTIVATE_POUT, HIGH); // HIGH = PSU disabled
    flushPSUActiviation(); // write active time to eeprom
    //psuIsDisabled = false;
  }
}

// writes active time to eeprom
void flushPSUActiviation() {
  addToCounter(EEP_TIM_PSU, (millis() - millisWhenPSUActivated) / 1000L );
  millisWhenPSUActivated = millis();
}

//
// =======================================================================================================
// APPLE CHIME START SOUND FSM
// =======================================================================================================
//

boolean melodyPlay;
boolean melodyIsPlaying = false;

void fsmChime() {

  static byte state = 1; // This variable is holding the current "step" of the FSM
  static unsigned long ts;  // To store the "current" time in for delays.
  static boolean wasInS5 = false;

  // if S3 State (Standby Mode) is active, reset bit "wasInS5"
  if (powerState == 3) {
    wasInS5 = false;
  }

  // if S5 State (OFF Mode) is active, set bit "wasInS5"
  if (powerState == 5) {
    wasInS5 = true;
  }

  switch (state)
  {

    //----Was Computer in S5 and is in S0 now?
    case 1:
      if ( wasInS5 && powerState == 0) { //delay 2.5s, then trigger Apple Chime
        // count system power (change from S5 to S0) cycles
        incrementCounter(EEP_CYC_POWER);
        state = 2;
      }
      else if  (melodyPlay) { // play chime immediately for testing
        Serial.println(F("Chime Test"));
        state = 4;
      }
      else state = 9;  // Don't play Chime
      break;

    //----Start timer
    case 2:
      ts = millis();  // Remember the current time
      state = 3;
      break;

    //----If time has passed, trigger Chime
    case 3:
      if (millis() > ts + eepromRead(EEP_CHIME_STARTUP_DELAY)) { // delay 2.5s
        state = 4;
      }
      break;

    //----Trigger Chime, reset bit "wasInS5"
    case 4:
      digitalWrite(CHIME_POUT, LOW);
      melodyIsPlaying = true;
      wasInS5 = false;
      state = 5;
      break;

    //----Start timer
    case 5:
      ts = millis();  // Remember the current time
      state = 6;
      break;

    //----If time has passed, delete Chime impulse
    case 6:
      if (millis() > ts + 100) { // Impulse length = 0.1s
        digitalWrite(CHIME_POUT, HIGH);
        melodyPlay = false;
        state = 7;
      }
      break;

    //----Start timer
    case 7:
      ts = millis();  // Remember the current time
      state = 8;
      break;

    //----If time has passed, delete "melodyIsPlaying" variable
    case 8:
      if (millis() > ts + eepromRead(EEP_CHIME_DURATION) ) { // delay about 2.8s to ensure, that the melody playback is finished
        // reason: fans should stay disabled during chime playback due to PWM noise in the speakers
        melodyIsPlaying = false;
        state = 9;
      }
      break;

    //----Start timer
    case 9:
      ts = millis();  // Remember the current time
      state = 10;
      break;

    //----If time has passed, go back to state 1
    case 10:
      if (millis() > ts + 100) {
        state = 1;
      }
      break;

    //----
    default:
      state = 1;
      break;
  }
}

//----Chime melody serial interface

// This function processes the fan related serial commands
void processCommandMelody(String subCmd) {

  noInterrupts();

  if (subCmd.equals("P")) {
    // Melody Play (Test function in the slider applet)
    melodyPlay = true;

  } else {
    Serial.println(F("Melody Command Unknown: MP (melody play)"));
  }
  interrupts();
}

//
// =======================================================================================================
// CCFL BACKLIGHT INVERTER
// =======================================================================================================
//

int inverterBrightness; // in %
boolean displaySleep; // true = display was put to sleep state with the "BD" serial command
unsigned long startupMillis;
unsigned long millsWhenLCDActivated = millis();
boolean StartDelayFinished = false;
boolean wasInS5 = true;
boolean brightnessWasChanged = false; // only allow writing the brightness to the EEPROM, if it was changed!
int brightnessChanged = 0;
int writeBrightCnt;

//----Inverter Setup

void setupInverter() {
  inverterEnabled = false;
  displaySleep = false;
  adjustInverter(eepromRead(EEP_BRIGHT)); // brightness after SMC reset in % (stored in the EEPROM)
}

//----Brightness

// This function increases the current brightness
void increaseBrightness () {
  adjustInverter(inverterBrightness = inverterBrightness + eepromRead(EEP_BRIGHT_INC));
}

// This function decreases the current brightness
void decreaseBrightness () {
  adjustInverter(inverterBrightness = inverterBrightness - eepromRead(EEP_BRIGHT_INC));
}

// This function limits and maps the inverter brightness and writes the value to the PWM output
// (called directly from Slider App (BWxx) or the Inc / Dec functions above)
void adjustInverter(int brightLevel) {
  
  static int oldBrightness;
  static int brightness;
  
  // Limit brightness range
  brightLevel = constrain (brightLevel, 0, 100); // limit brightness range
  inverterBrightness = brightLevel; // Report brightness back to slider App!!

  // Map brightness 0 - 100% equals to 140 - 0 PWM (0 PWM = full brightness!!)
  oldBrightness = brightness;
  brightness = map (brightLevel, 0, 100, eepromRead(EEP_MIN_BRIGHT), eepromRead(EEP_MAX_BRIGHT));

  if (inverterEnabled) {
    analogWrite(INVERTER_POUT, brightness);  // write new brightness
  }
  else {
    analogWrite(INVERTER_POUT, 255); // 255 = Backlights OFF
  }

  if (brightness != oldBrightness) {  // did brightness change?
    brightnessWasChanged = true;
    brightnessChanged++;
  }
}

//----Inverter

// This function enables / disables the inverter in accordance with the power state & the displaySleep state
void inverterOnOff() {
  if (powerState != 0 && inverterEnabled) { // disable inverter, if computer is NOT in S0 (ON) state & inverter is still enabled
    disableInverter();
    displaySleep = false;
  }
  if (powerState == 0 && ! displaySleep  && StartDelayFinished && ! inverterEnabled) { // enable inverter, if computer is in S0 state and display is NOT in "BD" sleep state
    enableInverter();
  }

  if (powerState == 0 && wasInS5) {
    startupMillis = millis();  // remember the time, when the computer changes from S5 (OFF) to S0 (ON)
    wasInS5 = false;
  }

  if (powerState == 5) {
    wasInS5 = true;  // remember, that the computer was in S5 (OFF) state
    StartDelayFinished = false;
  }

  if (powerState == 0 && millis() - startupMillis > eepromRead(EEP_INVERTER_STARTUP_DELAY)) {
    StartDelayFinished = true;  // after 6.5s enable the inverter (hide the NUC splash screen)
  }
}

// This function enables the Inverter and then writes the current brightness level to the PWM output
void enableInverter() {
  inverterEnabled = true;
  adjustInverter(inverterBrightness);
  millsWhenLCDActivated = millis();
}

// This function disables the Inverter and then writes the current brightness level to the PWM output
void disableInverter() {
  flushLCDActiviation();
  inverterEnabled = false;
  adjustInverter(inverterBrightness);
}

// This function writes the current inverter brightness to the EEPROM (triggered during shutdown)
void writeBrightToEEPROM() {
  if (brightnessWasChanged) {
    eepromWrite(EEP_BRIGHT, inverterBrightness);
    brightnessWasChanged = false;
    writeBrightCnt++;
  }
}

// This function writes active LCD hours to eprom
void flushLCDActiviation() {
  if (inverterEnabled) {
    addToCounter(EEP_TIM_LCD, (millis() - millsWhenLCDActivated) / 1000L );
    millsWhenLCDActivated = millis();
  }
}

//----Inverter serial interface

// This function processes the inverter related serial commands
void processCommandBrightness(String subCmd, String extraCmd) {

  noInterrupts();

  if (subCmd.equals("R")) {
    // Brightness Read (Slider Applet reads from the Arduino)
    Serial.println(inverterBrightness);

  } else if (subCmd.equals("+")) {
    // increase brightness
    increaseBrightness();

  } else if (subCmd.equals("-")) {
    // decrease brightness
    decreaseBrightness();

  } else if (subCmd.equals("W")) {
    // Brightness Write value BWnnn - nnn is the value (called, if slider position changes)
    adjustInverter( extraCmd.toInt() );

  } else if (subCmd.equals("A")) {
    // Activate Inverter
    displaySleep = false;
    enableInverter();
    serialCommandReceived();

  } else if (subCmd.equals("D")) {
    // Deactivate Inverter
    displaySleep = true;
    disableInverter();
    serialCommandReceived();

  } else {
    Serial.println(F("Brightness Command Unknown: BR (read), BW (write), BA (activate), BD (deactivate), B+, B- (change)"));
  }
  interrupts();
}

//
// =======================================================================================================
// CAPACITIVE TOUCH SENSORS
// =======================================================================================================
//

#ifdef CAPACITIVE

int upReading;
int dnReading;

void setupTouch() {
  void reset_CS_AutoCal();    // calibrate the touch sensors
  //upSensor.set_CS_AutocaL_Millis(0xBB8);     // autocalibrate timeout 3000ms  = HEX (0xBB8)
  //dnSensor.set_CS_AutocaL_Millis(0xBB8);
}

// Main Touch Sensor Reading
void touchControl() {

  //Read both sensors
  long start = millis();
  upReading = upSensor.capacitiveSensor(eepromRead(EEP_CAP_SAMPLE)); // (XX) is the number of sensor samples
  dnReading = dnSensor.capacitiveSensor(eepromRead(EEP_CAP_SAMPLE)); // about 80 = OK

  if (inverterEnabled && powerState == 0) { // Only allow brightness adjustment, if Computer is in S0 State

    // If the capacitive sensor reads above a certain threshold,
    if (upReading > eepromRead(EEP_UP_CAP_THR)) {  // about 70 = OK
      increaseBrightness();
    }
    if (dnReading > eepromRead(EEP_DN_CAP_THR)) {
      decreaseBrightness();
    }
  }
}

//----Touch sensor serial interface

// This function processes the touch sensor related serial commands
void processCommandCapacitive(String subCmd) {
  if (subCmd.equals("I")) {
    Serial.print(F("Down Sensor Reading = "));
    Serial.println(dnReading);
    Serial.print(F("Up Sensor Reading = "));
    Serial.println(upReading);

  } else {
    Serial.println(F("Capacative Command Unknown: CI (info)"));
  }
}

#endif

//
// =======================================================================================================
// FAN CONTROLLER FSM
// =======================================================================================================
//

volatile long fanRotationCount[8];  // "volatile" is required because of the interrupt functions!
static unsigned long fanTimeStart[8];
unsigned int fanRpm[8];
float sensTemp[3];

void fsmFans() {

  static byte state = 1; // This variable is holding the current "step" of the FSM
  static unsigned long ts;  // To store the "current" time in for delays.

  switch (state) {

    //----Wait
    case 1:
      //Do nothing
      ts = millis();  // Remember the current time
      state = 2;  // Move to the next state
      break;

    //----Wait for time
    case 2:
      // If time has passed, then move on to the next state.
      if (millis() > ts + 200) {
        state = 3;
      }
      break;

    //----Request temperature sensors for readings
    case 3:
      sensors.setWaitForConversion(false);  // (false) makes it async = no delay() is used inside the library
      sensors.requestTemperatures(); //Request temperatures (from all connected sensors)
      sensors.setWaitForConversion(true);
      ts = millis();  // Remember the current time
      state = 4;
      break;

    //----Wait for sensors (in accordance with sensor datasheet, required time depends on bit resolution!)
    case 4:
      // If time has passed, then go to state 6.
      if (millis() > ts + 800) {
        state = 5;
      }
      break;

    //----Read temperature sensors, write fan PWM
    case 5:
      // Read sensor 0, write fan 1 PWM (Case Fan)
      readTemperatureSensor(0, FAN1_CONTROL_POUT, FAN1_TACH_INT, eepromRead(EEP_MIN_FAN1_PWM), 255, eepromRead(EEP_FAN1_DIV)); //Sensor No, Fan No, Tach No, minPWM (30), maxPWM, Divider 500
      // Read sensor 1, write fan 2 PWM (HDD Fan)
      readTemperatureSensor(1, FAN2_CONTROL_POUT, FAN2_TACH_INT, eepromRead(EEP_MIN_FAN2_PWM), 255, eepromRead(EEP_FAN2_DIV)); //Sensor No, Fan No, Tach No, minPWM (20), maxPWM, Divider 750
      // Read sensor 2, write fan 3 PWM (CPU Fan)
      readTemperatureSensor(2, FAN3_CONTROL_POUT, FAN3_TACH_INT, eepromRead(EEP_MIN_FAN3_PWM), 255, eepromRead(EEP_FAN3_DIV)); //Sensor No, Fan No, Tach No, minPWM (30), maxPWM, Divider 500
      state = 1;
      break;

    //----
    default:
      state = 1;
      break;
  }
}

//Read temperature sensor, get fan RPM & write fan PWM function (called from FSM above)
void readTemperatureSensor(int sensorNo, int fanNo, int tachNo, int pwmMin, int pwmMax, int divider) {
  float sensorReading = (sensors.getTempCByIndex(sensorNo)); //Read temperature from sensor
  float pwmValue = sensorReading * sensorReading * sensorReading / divider;

  sensTemp[sensorNo] = sensorReading;

  if (getFanRPM(tachNo) < 100) {  //does fan spin? (tach < 100 means no)
    pwmMin = (pwmMin + pwmMax) / 2;  //Allows fan to start! 10 = OK for both small fans, big CPU fan does not start
  }

  pwmValue = constrain(pwmValue, pwmMin, pwmMax);  //Limit PWM range

  if (powerState != 0 || melodyIsPlaying == true || StartDelayFinished == false || (sensorReading < eepromRead(EEP_FAN3_TRESH) && fanNo == FAN3_CONTROL_POUT) ) {
    pwmValue = 0;  // Fan off, if computer is not in S0 (ON), during melody playback & during startup delay OR if Fan is CPU Fan & temperature is < 42 deg C
    // Reason: The CPU FAN is not necessary for cooling the NUC (during norma lperation) and stays completely off until the NUC exhaust air is >= 42 deg C.
  }

  analogWrite(fanNo, pwmValue); //Write Fan PWM
}

// This function calculates the current fan RPM
int getFanRPM(byte tachPin) {
  unsigned int rpm = fanRotationCount[tachPin] * 30000 / ( millis() - fanTimeStart[tachPin] );
  fanRotationCount[tachPin] = 0;
  fanTimeStart[tachPin] = millis();
  if (tachPin == FAN3_TACH_INT) {
    rpm = rpm / 2;
  }
  fanRpm[tachPin] = rpm;
  return  rpm;
}

// Interrupt functions for fan tach signals
void interrupFan1() {
  fanRotationCount[FAN1_TACH_INT] ++;
}

void interrupFan2() {
  fanRotationCount[FAN2_TACH_INT] ++;
}

void interrupFan3() {
  fanRotationCount[FAN3_TACH_INT] ++;
}

//----Fan serial interface

// This function processes the fan related serial commands
void processCommandFan(String subCmd) {

  if (subCmd.equals("R")) {
    // Fan Read (Slider Applet reads from the Arduino)
    Serial.print(fanRpm[FAN1_TACH_INT]);
    Serial.print(";");
    Serial.print(fanRpm[FAN2_TACH_INT]);
    Serial.print(";");
    Serial.print(fanRpm[FAN3_TACH_INT]);
    Serial.print(";");
    Serial.print(sensTemp[0]);
    Serial.print(";");
    Serial.print(sensTemp[1]);
    Serial.print(";");
    Serial.println(sensTemp[2]);

  } else {
    Serial.println(F("Fan Command Unknown: FR (read)"));
  }
}

//
// =======================================================================================================
// SLEEP LED CONTROLLER
// =======================================================================================================
//

byte attenuator = 0;
boolean ledTest = false;

//Slowly fade in or out LED attenuator in accordance with power state (called from Simple Timer every 10 ms)
void ledAttenuator() {

  // if S3 State (Standby Mode) or LED test is active or screen is dark, fade in and show LED
  if (powerState == 3 || ledTest || ! inverterEnabled) {
    attenuator = attenuator - 1;
  }
  // else fade LED fully out
  else {
    attenuator = attenuator + 1;
  }
  // Limit attenuator range
  attenuator = constrain(attenuator, 1, 255);
}

// LED controller
void ledSinus() {

  float brightness;
  byte ledMin = eepromRead(EEP_MIN_LED_PWM);
  byte ledMax = eepromRead(EEP_MAX_LED_PWM);

  if (!inverterEnabled && powerState == 0) {
    // LED always on, if screen is dark & computer is in S0 (ON) state
    brightness = ledMax / attenuator;
  }
  else if (powerState == 3 || ledTest) {
    // Breathing LED during system sleep (S3)
    float breathCycle = millis() / 4500.0; // 4500.0 means one sinus cycle per 4.5s
    float ledSin = (ledMax - ledMin) / 2;
    float ledOffset = ledMax - ledSin;
    // brightness = ( 134.0 + 121.0 * sin ( breathCycle * 2.0 * PI ) ) - attenuator;
    brightness = ( (ledOffset / (float)attenuator) + (ledSin / (float)attenuator) * sin (breathCycle * 2.0 * PI) - 1.7);
  }

  if (brightness <= 0.0) brightness = 0.0;
  analogWrite(SLEEP_LED_POUT, brightness); // breath LED during system sleep
}

//----LED serial interface

// This function processes the LED related serial commands
void processCommandLed(String subCmd) {

  if (subCmd.equals("T")) {
    ledTest = !ledTest;
    if (ledTest) {
      Serial.println(F("LED Test active -> type LT again"));
    } else {
      Serial.println(F("LED Test inactive"));
    }
  } else {
    Serial.println(F("LED Command Unknown: T (test)"));
  }
}

//
// =======================================================================================================
// DRIVE DOOR SERVO
// =======================================================================================================
//
#ifdef SERVO

boolean doorOpen;

void servoFunction() {
  static int pos;  //static variable holds servo position
  //byte servoSpeed = 3; // 1 - 6

  if (doorOpen == false) {
    //pos = 0;
    pos = doorServo.read() - eepromRead(EEP_DOOR_SPEED);  // speed range about 1 - 6

  } else {
    //pos = 180;
    pos = doorServo.read() + eepromRead(EEP_DOOR_SPEED);

  }
  pos = constrain(pos, eepromRead(EEP_DOOR_POS_CLOSED), eepromRead(EEP_DOOR_POS_OPENED));  //Limit servo position
  doorServo.write(pos);              // tell servo to go to position in variable 'pos'
}


//----Door serial interface

// This function processes the servo related serial commands
void processCommandDoor(String subCmd) {

  if (subCmd.equals("S")) {
    doorOpen = !doorOpen;
    if (doorOpen) {
      Serial.println(F("Door opened"));
    } else {
      Serial.println(F("Door closed"));
    }
  } else {
    Serial.println(F("Door Command Unknown: S (switch)"));
  }
}

#endif

//
// =======================================================================================================
// USB SERIAL CONTROLLER
// =======================================================================================================
//

// PROCESS INPUT
void loopSerialCommandProcessor() {

  static char serData[32]; // Buffer
  static byte count = 0;
  static boolean initialised = false;

  if (! initialised) {
    // fist tie into this method setup
    Serial.begin(SERIAL_BAUD_RATE);
    initialised = true;
  }

  // if serial port not ready (Leonardo only).
  if (!Serial) return; // cant do anything

  while (Serial.available() > 0) {

    // read a byte and add to the buffer
    char inByte = Serial.read();

    // if the read failed
    if (inByte == -1) {
      break;

      // if EOL (End Of Line)
    } else if ( inByte == CR || inByte == LF ) {

      // terminate the buffer
      serData[count++] = 0;

      //Convert The array to a String
      String command = String(serData);

      if (command.length() > 0 ) {

        // process the command
        processCommand(command);
      }

      // clear the command buffer
      count = 0;

      // protect against overrun of buffer
    } else if (count < 30) {
      serData[count++] = inByte;
    }
  }
}

// Main function to process a serial command from terminal or Brightness Menulet
void processCommand(String cmd) {

  String firstCmd = cmd.substring(0, 1);
  String secondCmd = cmd.substring(1, 2);
  String extraCmd = cmd.substring(2);

  if (firstCmd.equals("B")) processCommandBrightness(secondCmd, extraCmd);

#ifdef CAPACITIVE
  else if (firstCmd.equals("C")) processCommandCapacitive(secondCmd);
#endif

#ifdef SERVO
  else if (firstCmd.equals("D")) processCommandDoor(secondCmd);
#endif

  else if (firstCmd.equals("E")) processCommandEeprom(secondCmd, extraCmd);

  else if (firstCmd.equals("L")) processCommandLed(secondCmd);

  else if (firstCmd.equals("M")) processCommandMelody(secondCmd);

  else if (firstCmd.equals("F")) processCommandFan(secondCmd);

  else if (firstCmd.equals("S")) processCommandSystem(secondCmd);

#if defined(CAPACITIVE) && defined(SERVO)
  else Serial.println(F("Command Unknown: B (brightness), C (capacitive), D (door), E (eeprom), F (fans), L (LED), M (melody), S (system)"));
#endif

#if defined(CAPACITIVE) && !defined(SERVO)
  else Serial.println(F("Command Unknown: B (brightness), C (capacitive), E (eeprom), F (fans), L (LED), M (melody), S (system)"));
#endif

#if !defined(CAPACITIVE) && defined(SERVO)
  else Serial.println(F("Command Unknown: B (brightness), D (door), E (eeprom), F (fans), L (LED), M (melody), S (system)"));
#endif

#if !defined(CAPACITIVE) && !defined(SERVO)
  else Serial.println(F("Command Unknown: B (brightness), E (eeprom), F (fans), L (LED), M (melody), S (system)"));
#endif
}


// This function sends "OK" back to the Brightness Menulet (only intended for critical commands)
void serialCommandReceived() {
  Serial.println("OK");
}

//
// =======================================================================================================
// SYSTEM STATISTICS
// =======================================================================================================
//

void processCommandSystem(String subCmd) {

  // Statistics
  if (subCmd.equals("S")) {

    flushPSUActiviation(); // Flushes active PSU Hours to EEPROM
    flushLCDActiviation(); // Flushes LCD Backlight Hours to EEPROM

    Serial.print(F("Mains Power Cycles: "));
    Serial.println(eepromRead(EEP_CYC_MAINS));

    Serial.print(F("Power On Cycles: "));
    Serial.println(eepromRead(EEP_CYC_POWER));

    Serial.print(F("Sleep Cycles: "));
    Serial.println(eepromRead(EEP_CYC_SLEEP));

    Serial.print(F("PSU Enabled Hours: "));
    printHours(eepromRead(EEP_TIM_PSU));
    Serial.println();

    Serial.print(F("Active (S0) Hours: "));
    printHours(eepromRead(EEP_TIM_PSU) - eepromRead(EEP_TIM_SLEEP));
    Serial.println();

    Serial.print(F("LCD Backlight Hours: "));
    printHours(eepromRead(EEP_TIM_LCD));
    Serial.println();

    Serial.print(F("Number of EEPROM entries: "));
    Serial.println(eepromRead(EEP_VERSION));

    // System Timers in use
    Serial.print(F("SimpleTimer threads used: "));
    Serial.print(timer.getNumTimers());
    Serial.println(F(" of 10"));

    // System Ram Available in bytes. Leonardo has 2.5Kbyes
    Serial.print(F("RAM in use: "));
    Serial.print(2560L - freeRam());
    Serial.println(F(" of 2560 bytes"));

    // Uptime
  } else if (subCmd.equals("U")) {

    Serial.println(F("Readings for current SMC Power Cycle:"));
    Serial.println();

    Serial.print(F("SMC Uptime: "));
    printTime(millis() / 1000L);
    Serial.println();

    Serial.print(F("CPU Uptime: "));
    printTime((millis() - millisOfCpuStartup) / 1000L);
    Serial.println(F(" (current S0 cycle)"));

    Serial.print(F("Number of Brightness changes: "));
    Serial.println(brightnessChanged);

    Serial.print(F("Number of Brightness EEPROM writes: "));
    Serial.println(writeBrightCnt);

    Serial.print(F("Number of total EEPROM writes: "));
    Serial.println(epromWriteCnt);

    // Temperature
  } else if (subCmd.equals("T")) {

    Serial.print(F("Case:   "));
    Serial.print(fanRpm[FAN1_TACH_INT]);
    Serial.print(F(RPM));
    Serial.print(sensTemp[0]);
    Serial.println(DEGC);

    Serial.print(F("HDD:   "));
    Serial.print(fanRpm[FAN2_TACH_INT]);
    Serial.print(F(RPM));
    Serial.print(sensTemp[1]);
    Serial.println(DEGC);

    Serial.print(F("CPU:   "));
    Serial.print(fanRpm[FAN3_TACH_INT]);
    Serial.print(F(RPM));
    Serial.print(sensTemp[2]);
    Serial.println(DEGC);

  } else {
    Serial.println(F("System Command Unknown: S (stats), T (temps), U (uptime)"));
  }
}

// Print Time is for SU
void printTime(unsigned long secs) {

  unsigned long mins = secs / 60L;
  secs = secs - (mins * 60);
  unsigned long hrs = mins / 60L;
  mins = mins - (hrs * 60);
  int days = hrs / 24;
  hrs = hrs - (days * 24);

  if (days > 0) {
    Serial.print(days);
    Serial.print(F("d "));
  }

  if (hrs > 0) {
    Serial.print(hrs);
    Serial.print(F("h "));
  }

  if (mins > 0) {
    Serial.print(mins);
    Serial.print(F("m "));
  }

  Serial.print(secs);
  Serial.print(F("s "));

}


// Print Hours is for SS
void printHours(unsigned long secs) {

  unsigned long hrs = secs / 3600;
  unsigned long decimal = (secs - (hrs * 3600)) / 360L;

  Serial.print(hrs);
  Serial.print(".");
  Serial.print(decimal);
}


//
// =======================================================================================================
// GENERAL PURPOSE UTILITY FUNCTIONS
// =======================================================================================================
//

// Function to return the free SRAM in bytes between heap and stack
// https://learn.adafruit.com/memories-of-an-arduino/measuring-free-memory

int freeRam () {
  extern int __heap_start, *__brkval;
  int v;
  return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}
